"""File-related utility functions.

Implements: `load_file`, `dump_file`, `load_json_file`,
`dump_json_file`, `get_parent_dir_path`, `get_dir_checksum`,
`get_file_checksum`, `rel_to_abs_path`, `expect_file_contents`"""

from __future__ import annotations
from typing import Any, List, Optional
import traceback
import hashlib
import json
import os
import tum_esm_utils


def load_file(path: str) -> str:
    with open(path, "r") as f:
        return f.read()


def dump_file(path: str, content: str) -> None:
    with open(path, "w") as f:
        f.write(content)


def load_json_file(path: str) -> Any:
    with open(path, "r") as f:
        return json.load(f)


def dump_json_file(path: str, content: Any, indent: Optional[int] = 4) -> None:
    with open(path, "w") as f:
        json.dump(content, f, indent=indent)


def get_parent_dir_path(script_path: str, current_depth: int = 1) -> str:
    """Get the absolute path of a parent directory based on the
    current script path. Simply pass the `__file__` variable of
    the current script to this function. Depth of 1 will return
    the direct parent directory of the current script."""

    assert current_depth > 0, "depth must be greater than 0"
    output = os.path.dirname(os.path.abspath(script_path))
    for _ in range(current_depth - 1):
        output = os.path.dirname(output)
    return output


def get_dir_checksum(path: str) -> str:
    """Get the checksum of a directory using md5deep."""

    assert os.path.isdir(path), f"{path} is not a directory"
    return tum_esm_utils.shell.run_shell_command(
        f"md5deep -r -b {path} | sort -u | md5sum"
    )


def get_file_checksum(path: str) -> str:
    """Get the checksum of a file using MD5 from `haslib`.

    Significantly faster than `get_dir_checksum` since it does
    not spawn a new process."""

    assert os.path.isfile(path), f"{path} is not a file"
    with open(path, "rb") as f:
        return hashlib.md5(f.read()).hexdigest()


def rel_to_abs_path(*path: str) -> str:
    """Convert a path relative to the caller's file to an absolute path.
    
    Inside file `/home/somedir/somepath/somefile.py`, calling
    `rel_to_abs_path("..", "config", "config.json")` will
    return `/home/somedir/config/config.json`.
    
    Credits to https://stackoverflow.com/a/59004672/8255842"""

    return os.path.normpath(
        os.path.join(
            os.path.dirname(traceback.extract_stack()[-2].filename),
            *path,
        )
    )


def read_last_n_lines(
    file_path: str,
    n: int,
    ignore_trailing_whitespace: bool = False,
) -> List[str]:
    """Read the last `n` lines of a file.

    The function returns less than `n` lines if the file has less than `n` lines.
    The last element in the list is the last line of the file.
    
    This function uses seeking in order not to read the full file. The simple
    approach of reading the last 10 lines would be:

    ```python
    with open(path, "r") as f:
        return f.read().split("\\n")[:-10]
    ```

    However, this would read the full file and if we only need to read 10 lines
    out of a 2GB file, this would be a big waste of resources.
    
    The `ignore_trailing_whitespace` option to crop off trailing whitespace, i.e.
    only return the last `n` lines that are not empty or only contain whitespace."""

    with open(file_path, "rb") as f:
        f.seek(-1, os.SEEK_END)

        if ignore_trailing_whitespace:
            while f.read(1) in [b"\n", b" ", b"\t"]:
                try:
                    f.seek(-2, os.SEEK_CUR)
                except OSError:
                    # reached the beginning of the file
                    return [""]

            f.seek(-1, os.SEEK_CUR)
            # now the cursor is right before the last
            # character that is not a newline or a space

        last_characters: bytes = b""
        new_line_chars_seen: int = 0

        while True:
            try:
                new_character = f.read(1)
                if new_character == b"\n":
                    new_line_chars_seen += 1
                if new_line_chars_seen == n:
                    break
                last_characters += new_character
                f.seek(-2, os.SEEK_CUR)
            except OSError:
                # reached the beginning of the file
                break

    return last_characters.decode()[::-1].split("\n")


def expect_file_contents(
    filepath: str,
    required_content_blocks: list[str] = [],
    forbidden_content_blocks: list[str] = [],
) -> None:
    """Assert that the given file contains all of the required content
    blocks, and/or none of the forbidden content blocks.
    
    Args:
        filepath:                 The path to the file.
        required_content_blocks:  A list of strings that must be present in the file.
        forbidden_content_blocks: A list of strings that must not be present in the file.
    """

    with open(filepath, "r") as f:
        file_content = f.read()

    for b in required_content_blocks:
        assert b in file_content, f'required log content block not found "{b}"'

    for b in forbidden_content_blocks:
        assert b not in file_content, f'forbidden log content block found "{b}"'



"""Better defaults for matplotlib plots and utilities for creating and saving figures.

Implements: `apply_better_defaults`, `create_figure`, `add_subplot`

This requires you to install this utils library with the optional `plotting` dependencies:

```bash
pip install "tum_esm_utils[plotting]"
# or
pdm add "tum_esm_utils[plotting]"
```"""

from typing import Generator, Optional, Any
import contextlib

import matplotlib.pyplot as plt
import matplotlib.font_manager
from tailwind_colors import TAILWIND_COLORS_HEX


def apply_better_defaults(font_family: Optional[str] = "Roboto") -> None:
    """Apply better defaults to matplotlib plots.

    Args:
        font_family: The font family to use for the plots. If None, the default
                        settings are not changed.
    """

    if font_family is not None:
        system_fonts = matplotlib.font_manager.findSystemFonts()
        matching_fonts = [font for font in system_fonts if font_family in font]
        if len(matching_fonts) == 0:
            raise ValueError(
                f"Font family '{font_family}' not found. System fonts: {system_fonts}"
            )
        for font in matching_fonts:
            matplotlib.font_manager.fontManager.addfont(font)
        plt.rcParams['font.sans-serif'] = [
            font_family, *plt.rcParams['font.sans-serif']
        ]

    plt.rcParams['figure.titleweight'] = 'bold'
    plt.rcParams['axes.titleweight'] = 'semibold'
    plt.rcParams['axes.labelweight'] = 'semibold'
    plt.rcParams['axes.facecolor'] = TAILWIND_COLORS_HEX.SLATE_050
    plt.rcParams['axes.edgecolor'] = TAILWIND_COLORS_HEX.SLATE_600
    plt.rcParams['grid.color'] = TAILWIND_COLORS_HEX.SLATE_300
    plt.rcParams['axes.axisbelow'] = True
    plt.rcParams['xtick.color'] = TAILWIND_COLORS_HEX.SLATE_600
    plt.rcParams['ytick.color'] = TAILWIND_COLORS_HEX.SLATE_600
    plt.rcParams['xtick.labelcolor'] = "black"
    plt.rcParams['ytick.labelcolor'] = "black"
    plt.rcParams['scatter.edgecolors'] = "none"
    matplotlib.style.use('fast')


@contextlib.contextmanager
def create_figure(
    path: str,
    title: Optional[str] = None,
    width: float = 10,
    height: float = 10,
    suptitle_y: float = 0.97,
    padding: float = 2,
    dpi: int = 250,
) -> Generator[plt.Figure, None, None]:
    """Create a figure for plotting.
    
    Usage:

    ```python
    with create_figure("path/to/figure.png", title="Title") as fig:
        ...
    ```
    Args:
        path: The path to save the figure to.
        title: The title of the figure.
        width: The width of the figure.
        height: The height of the figure.
        suptitle_y: The y-coordinate of the figure title.
        padding: The padding of the figure.
        dpi: The DPI of the figure.
    """

    fig = plt.figure()
    fig.set_size_inches(width, height)
    yield fig
    if title is not None:
        fig.suptitle(title)
        fig.suptitle(fig._suptitle.get_text(), y=suptitle_y)  # type: ignore
    fig.tight_layout(pad=padding)
    fig.savefig(path, dpi=dpi)
    plt.close(fig)


def add_subplot(
    fig: plt.Figure,
    position: tuple[int, int, int],
    title: Optional[str] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    **kwargs: dict[str, Any],
) -> plt.Axes:
    """Add a subplot to a figure.
    
    Args:
        fig: The figure to add the subplot to.
        position: The position of the subplot. The tuple should contain three integers: the number of rows, the number of columns, and the index of the subplot.
        title: The title of the subplot.
        xlabel: The x-axis label of the subplot.
        ylabel: The y-axis label of the subplot.
        **kwargs: Additional keyword arguments for the subplot.
    
    Returns:
        An axis object for the new subplot.
        
    Raises:
        ValueError: If the index of the subplot is invalid."""

    if (position[2] < 1) or (position[2] > (position[0] * position[1])):
        raise ValueError(
            "Invalid subplot index. The index must be between 1 and the number of rows times the number of columns."
        )

    axis = fig.add_subplot(position[0], position[1], position[2], **kwargs)
    if title is not None:
        axis.set_title(title)
    if xlabel is not None:
        axis.set_xlabel(xlabel)
    if ylabel is not None:
        axis.set_ylabel(ylabel)
    axis.grid(visible=None, which="minor", axis="both", linewidth=0.25)
    axis.grid(visible=None, which="major", axis="both", linewidth=0.75)
    return axis



"""Python utilities by the Professorship of Environmental
Sensing and Modeling at the Technical University of Munich.

GitHub Repository https://github.com/tum-esm/utils
Documentation: https://tum-esm-utils.netlify.app
PyPI: https://pypi.org/project/tum-esm-utils"""

from . import (
    code,
    datastructures,
    decorators,
    files,
    mathematics,
    processes,
    shell,
    system,
    text,
    timing,
    validators,
)

# ignore import errors from the following submodules
# because they requires extras to be installed

# requires extra "polars"
try:
    from . import em27
except ImportError:
    pass

# requires extra "plotting"
try:
    from . import plotting
except ImportError:
    pass
